<?php
/**
 * About routing
 * The goals of routing is how to find the controller and action.
 * Assume the request url as following:
 * http://localhost/a/b
 *
 * 1. Generally,
 * controller path: /controllers/
 * controller name: 'AController' at acontroller.php
 * action name: 'b()'
 *
 * 2. If defined 'a' in gModules().
 * controller path: /controllers/a/
 * controller name: 'BController' at bcontroller.php
 * action name: '' but auto-match to 'index()' // index() is default action.
 */

/**
 * Router base class
 * @package base
 * @since 1.0
 * @author JJB <splitedragon@hotmail.com>
 * @copyright Copyright (c) 2019, EVENING
 */
class Router
{
    private static $_modules = null;
    private static $_routes = null;
    private static $_url = array();

    public static function getUrl($key=NULL)
    {
        return ($key == NULL)? self::$_url : self::$_url[$key];
    }

    public static function setUrl($key, $val)
    {
        self::$_url[$key] = $val;
    }


    public static function getRequestURL()
    {
        // Added by jjb 2020.04.24. For CLI version.
        if(gIsCLI() && !isset($_SERVER["REQUEST_URI"])) {
            $args = $_SERVER['argv'];//array_slice($_SERVER['argv'], 1);
            $_SERVER["REQUEST_URI"] = $args ? implode('/', $args) : '';
        }
        // End adding

        $request = strtok($_SERVER["REQUEST_URI"],'?');
        $request = self::clearURL($request); // Always starting from '/'
        $request = substr($request, strlen(BASE_URL));
        $request = preg_replace('/^\\/index.php\\//', '', $request.'/', 1); // Modified by jjb
        $request = self::clearURL($request); // Always start from '/' to end with no '/'

        return $request;
    }
    /**
     * Analyze the requested URL.
     * Assume defined 'admin' as module in gModules()!
     * e.g. Client request to 'http://localhost/admin/a/b/c?a=123&b=456' or 'http://localhost/index.php/admin/a/b/c?a=123&b=456p'
     *  array (size=7)
     *  'request' => string 'a/b/c'
     *  'module' => string 'admin'
     *  'path' => string 'admin'
     *  'controller' => string 'A'
     *  'action' => string 'b'
     *  'argument' => array ( 0 => string 'c' )
     *  'http' => string 'http://::1/evening/admin/a/b/c?a=123&b=456'
     *
     * Default is based on $g_default_router of /config/url.config.php
     *
     * @return array Analyzed request url array.
     */
    public static function prepare()
    {
        $path = '';
        $definedModule = '';
        $definedRoute = null;
        $controller = null;
        $action = null;
        $model = null;
        $argument = null;
        $default_router = gDefaultRoute();

        $request = self::getRequestURL(); // Always starting from '/'

        $definedRoute = self::inRoutes($request);

        if(!empty($definedRoute) && is_array($definedRoute)) {
            $key = $definedRoute['key'];
            $rule = $definedRoute['rule'];
            if($key == '*') {
                $controller = isset($rule[0])? $rule[0] : null;
                $action = isset($rule[1])? $rule[1] : null;
                $argument = null;
            } else {
                $m = $request==$key? '' : substr($request, strlen($key));
                $arr = explode('/', $m);
                $arr = self::removeEmptyItem($arr);
                $controller =   isset($rule[0])? $rule[0] : null;
                if($controller != null) {
                    if(isset($rule[1])) {
                        $action = $rule[1];
                        $argument = empty($arr)? null : $arr;
                    } else {
                        $action = isset($arr[0])? $arr[0] : null;
                        @array_shift($arr);
                        $argument = empty($arr)? null : $arr;
                    }
                }
            }
            $path = isset($rule[2])? $rule[2] : '';
        }
        else {
            $b = explode('/', $request);
            // Check modules!
            $definedModule = self::inModules($request);

            if ($definedModule != '') {
                $m = substr($request, strlen($definedModule));
                $b = explode('/', self::clearURL($m));
            }
            $b = self::removeEmptyItem($b);
            $size = count($b);

            if($size>0) { // Request is now including at least a controller.
                // Detect arguments.
                if($size > 2) { // At least one or more arguments were included in the request.
                    $argument = array_slice($b, 2);
                }

                $controller = isset($b[0]) ? $b[0] : null;
                $action = isset($b[1]) ? $b[1] : null;
            }

            if($definedModule != '') {
                $path = $definedModule.$path;
            }
        }

        $controller = empty($controller)? $default_router['controller'] : $controller;
        $action = empty($action)? $default_router['action'] : $action;

        // Commented by JJB 2020.12.14.
        /*if (!preg_match("/^[a-z0-9_]+$/i", $controller) && !empty($controller)) {
            $page = new Errors('404');
            $page->message = 'Invalid controller: '.Filter::safeString($controller);
            $page->show();
            $controller = null;
        }
        if (!preg_match("/^[a-z0-9_]+$/i", $action) && !empty($action)) {
            $page = new Errors('404');
            $page->message = 'Invalid action: '.Filter::safeString($action);
            $page->show();
            $action = null;
        }*/

        $model = ucwords($controller);
        $controller = ucwords($controller);

        $url = array();
        $url['request'] = Filter::safeString($request);
        $url['defined_module'] = Filter::safeString($definedModule);
        $url['defined_route'] = $definedRoute;
        $url['path'] = Filter::safeString($path);
        $url['controller'] = Filter::safeString($controller);
        $url['model'] = Filter::safeString($model);
        $url['action'] = Filter::safeString($action);
        $url['argument'] = $argument;

        // Modified by jjb 2020.04.24. For CLI version.
        if(gIsCLI()) {
            $url['http'] = $_SERVER['REQUEST_URI'];
        } else {
            $url['http'] = (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] === 'on' ? "https" : "http") . "://" . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'];
        }
        // End modifying


        self::setUrl('http', $url['http']);
        self::setUrl('request', $url['request']);
        self::setUrl('defined_module', $url['defined_module']);
        self::setUrl('defined_route', $url['defined_route']);
        self::setUrl('path', $url['path']);
        self::setUrl('controller', $url['controller']);
        //self::setUrl('model', $url['model']);
        self::setUrl('action', $url['action']);
        self::setUrl('argument', $url['argument']);
        //self::setUrl('GET', $_GET);
        //self::setUrl('POST', $_POST);
        //var_dump(self::getUrl());
        return self::getUrl();
    }

    public static function redirect($url)
    {
        @header("Location: ".Filter::toStr($url));
        exit;
    }

    public static function reload()
    {
        //header("Location: ".$_SERVER['PHP_SELF']);
        @header("Refresh: 0");
        exit;
    }

    public static function backward()
    {
        echo "<script>window.history.back();</script>";
    }

    public static function forward()
    {
        echo "<script>window.history.forward();</script>";
    }

    public static function controller($val)
    {
        $path = trim($val, " \t\n\r\0\x0B\\/");
        $path = str_replace(array('\\', '/'), array(DS, DS), $path);
        $a = explode(DS, $path);
        $size = count($a);
        if(!isset($a[$size-1]) || empty($a[$size-1])) {
            $err = new Errors(500);
            $err->message = 'Invalid controller: '.$val;
            $err->show();
        }

        $controllerName = ucwords($a[$size-1]).'Controller';
        array_splice($a, $size-1);
        $a = self::removeEmptyItem($a);
        $path = implode(DS, $a);

        //$fileName = strtolower($controllerName).'.php';
        $fileName = $controllerName.'.php'; // For case-sensitive of UNIX

        $file = gURL('controllers').DS.$path;
        $file .= $path!=''? DS.$fileName : $fileName;
        return array('filePath'=>$file, 'className'=>$controllerName);
    }

    public static function model($val)
    {
        $path = trim($val, " \t\n\r\0\x0B\\/");
        $path = str_replace(array('\\', '/'), array(DS, DS), $path);
        $a = explode(DS, $path);
        $size = count($a);
        if(!isset($a[$size-1]) || empty($a[$size-1])) {
            $err = new Errors(500);
            $err->message = 'Invalid model: '.$val;
            $err->show();
        }

        $modelName = ucwords($a[$size-1]).'Model';
        array_splice($a, $size-1);
        $a = self::removeEmptyItem($a);
        $path = implode(DS, $a);

        //$fileName = strtolower($modelName).'.php';
        $fileName = $modelName.'.php'; // For case-sensitive of UNIX

        $file = gURL('models').DS.$path;
        $file .= $path!=''? DS.$fileName : $fileName;
        return array('filePath'=>$file, 'className'=>$modelName);
    }

    public static function view($val)
    {
        $path = trim($val, " \t\n\r\0\x0B\\/");
        $path = str_replace(array('\\', '/'), array(DS, DS), $path);
        $a = explode(DS, $path);
        $size = count($a);
        if(!isset($a[$size-1]) || empty($a[$size-1])) {
            $err = new Errors(500);
            $err->message = 'Invalid view: '.$val;
            $err->show();
        }

        //$viewName = ucwords($a[$size-1]);
        $viewName = $a[$size-1];
        array_splice($a, $size-1);
        $a = self::removeEmptyItem($a);
        $path = implode(DS, $a);

        //$fileName = strtolower($viewName).'.php';
        $fileName = $viewName.'.php'; // For case-sensitive of UNIX

        $file = gURL('views').DS.$path;
        $file .= $path!=''? DS.$fileName : $fileName;

        return array('filePath'=>$file, 'className'=>$viewName);
    }

    protected static function parse($val)
    {
        $result = null;
        if(!is_array(self::$_routes)) {
            self::$_routes = gRoutes();
        }

        if(isset(self::$_routes[$val]) && !empty(self::$_routes[$val]) && is_array(self::$_routes[$val])) {
            $result = self::$_routes[$val];
            if(count($result) < 2) {
                $result[1] = 'index';
            }
        }
        return $result;
    }

    protected static function inRoutes($val)
    {
        $method = isset($_SERVER['REQUEST_METHOD'])? strtoupper($_SERVER['REQUEST_METHOD']) : 'CLI'; // GET, POST, PUT, DELETE

        $default = gDefaultRoute();

        $result = null;

        if(!is_array(self::$_routes)) {
            self::$_routes = gRoutes();
        }

        $routes = isset(self::$_routes[$method])? self::$_routes[$method] : array();

        $key = NULL;
        foreach ($routes as $k=>$item) {
            $b = self::clearURL($k);
            $pos = strpos($val.'/', $b.'/');

            if($pos === 0) {
                $result = $item;
                $key = $b;
                break;
            }
        }
        if($key==NULL || $result==NULL ) {
            if (array_key_exists('*', $routes)) {
                $result = $routes['*'];
                $key = '*';
            } else {
                return NULL;
            }
        }

        return array('key'=>$key, 'rule'=>$result);
    }

    protected static function inModules($val)
    {
        $module = '';
        if(!is_array(self::$_modules)) {
            self::$_modules = gModules();
        }

        $modules = self::$_modules;

        $a = self::clearURL($val);

        for($i=0; $i<count($modules); $i++){
            $b = self::clearURL($modules[$i]);
            if(strpos($a.'/', $b.'/') === 0) {
                $module = $b;
                break;
            }
        }

        return $module;
    }

    protected static function removeEmptyItem($arr)
    {
        return array_values(array_filter($arr, function($item) {
            return $item!='';
        }));
    }

    /**
     * Fixing slash in path
     * /a/////b//////c/ changed to /a/b/c
     * i.e. pre-slash and end-slash are removed, also remove empty slash like /////b->/b
     * @param $val
     * @return string
     */
    protected static function clearURL($val)
    {
        /*$a = explode('/', $val);
        $a = self::removeEmptyItem($a);
        $result = implode('/', $a);
        return $result;*/

        $result = trim($val, " \t\n\r\0\x0B/");
        $result = preg_replace('/\\/\\/+/', '/', $result);
        return '/'.$result;
    }
}